/*
 * Copyright 2005, 2006 University of Leiden.
 *
 * This file is part of MIA+EWS.
 *
 * MIA+EWS is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MIA+EWS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MIA; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/*******************************************************************************
* MPIA, NEVEC - MIDI project
*
* "@(#) $Id: simulate.c,v 1.7 2002/06/04 09:29:29 jdejong Exp $"
*
* who        when      what
* ---------  --------  ----------------------------------------------
* jdejong    2002-05-08 imported
*
*/
/* $Id: simulate.c,v 1.7 2002/06/04 09:29:29 jdejong Exp $ */
/*----------------------------------------------------------------------- */
/* Routines and parameters for midi data simulation                       */
/*----------------------------------------------------------------------- */
#include "simulate.h"
#include "interpol.h"

/* internal prototypes */
void OPDfile_next (FILE *hFile, double *_time, float *delay);
void OPDfile_find (FILE *hFile, double _time);

/* file local stuff for data simulation */
/* flag to indicate active and successful reading of OPD file */
static int  IamOK = 0;  
/* arrays for interpolating OPDs from a file */
static double OPDtime[3] = {0.0, 0.0, 0.0}; /* times in OPDdelay */
static float OPDdelay[3] = {0.0, 0.0, 0.0}; /* delay at OPDtime */
/* interpolation kernal */
static float OPDintr[10]={0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0}; 
static int   OPDcen=1;    /* first (1-rel) value to use */
static float OPDscale=1.0; /* scaling for tau0 */
/* has the pixel range been blown before */
static float max_badfact=0.0;

imaging_data_table_def* 
make_data_table (fitsfile *fptr, int nx, int ny, int nreg, int maxstep, 
		 int *ierr)
/*----------------------------------------------------------------------- */
/*   Create FITS file and initialize table                                */
/*   Inputs:                                                              */
/*      fptr      cfitsio pointer to open FITS file                       */
/*      nx        Size of data array in x direction                       */
/*      ny        Size of data array in y direction                       */
/*      nreg      Number of regions                                       */
/*      maxstep   Maximum number of steps in obs. sequence.               */
/*   Output:                                                              */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*   Returns:                                                             */
/*       imaging_data_table_def structure to use to write table           */
/*       NULL on failure                                                  */
/*----------------------------------------------------------------------- */
{
  imaging_data_table_def *out;
  imaging_detector_table_def *detector;
  int ireg;
  char *instrument="simulator";

  if (*ierr>0) return NULL; /* existing error condition? */

  /* make simple imaging_detector_table_def */
  detector = make_testing_detector_table_def(nx, ny, nreg);
  if (detector==NULL) { /* error check */
    *ierr = 10000;
    return NULL;
  } /* end error test */

  /* modify imaging_detector_table_def */
  /* for now use regular frequency spacing */
  for (ireg=0; ireg<nreg; ireg++) { /* per region */
    detector->gain[ireg] = DetGain;
    /* first axis is frequency */
    strcpy (detector->ctype[ireg][0], "FREQ    ");
    /* frequency at reference pixel */
    detector->crval[ireg][0] = CLIGHT / (Wavelength * 1.0e-6);
    /* reference pixel */
    detector->crpix[ireg][0] = 1.5 + Nchan * 0.5;
    /* frequency spacing */
    detector->cd[ireg][0] =  
      ((CLIGHT / ((Wavelength - 0.5 * DW) * 1.0e-6)) -
       (CLIGHT / ((Wavelength + 0.5 * DW) * 1.0e-6))) / Nchan;
  }

  /* name of instrument */
  if (detector->instrument) free (detector->instrument);
  detector->instrument = (char*)malloc(strlen(instrument)+1);
  strcpy (detector->instrument, instrument);

  /* write imaging_detector table and open/init imaging data table */
  out = make_imaging_data_table_def(fptr, detector, maxstep, ierr);
  if (out==NULL) { /* failed */
    *ierr = 10000;
    kill_imaging_detector_table_def(detector); /* clean up */
    return NULL;
  } /* end error test */

  /* number of steps */
  out->maxstep = maxstep;

  /* add parameters in history NO SUCH FUNCTION */
  /* history(fptr, ierr); */

  return out; /* done */
} /* end make_data_table */

void simulate_data (imaging_data_row *row, float OPD_off,
		    float amplitude, int integ, float delay, 
		    float rate, float rmsOPD, float *gain, float *bg,
		    imaging_detector_table_def* detector)
/*----------------------------------------------------------------------- */
/*   Simulate telescope data                                              */
/*   Inputs:                                                              */
/*      row       row data structure to be filled in                      */
/*      OPD_off   OPD error for this sample                               */
/*      amplitude Amplitude of correlated signal in mJy                   */
/*      integ     integration time in msec                                */
/*      delay     Delay to add (microns)                                  */
/*      rate      Delay rate to add (microns/sec)                         */
/*      rmsOPD    RMS OPD noise per sample                                */
/*      gain      frequency gain factors for source flux                  */
/*      bg        background factor per frequency channel                 */
/*      detector  imaging_detector table definition structure             */
/*----------------------------------------------------------------------- */
{
  long i, j, ireg, indx;
  double freq, freq0, freq_incr, delay_off, time_sec;
  float  ref_pix, noise, background, value, bgcorr, phase, amp;
  float  arg, bw_decorr, corr_amp;
  int nx, ny, nreg;
  int badrange=0;
  float badfact=1.0, testfact;

  /* get info from structures */
  nx        = detector->naxis[0][0];    /* number of pixels in Freq */
  ny        = detector->naxis[0][1];    /* number of pixels in Y */
  nreg      = detector->nregion;        /* number of regions */
  freq0     = detector->crval[0][0]; /* frequency at reference pixel */
  ref_pix   = detector->crpix[0][0]; /* reference pixel 1-rel */
  freq_incr = detector->cd[0][0];    /* frequency spacing */
  time_sec = row->time * 86400.0;    /* time in seconds */
  amp = (amplitude * (Counts*integ*0.001)) * Coher * Fract / Nchan;

  background = Fract * Sky * integ / Nchan;   /* Background noise */
  noise = sqrt(background + RON*RON); /* shot plus readout noise */
  delay_off = delay + time_sec * rate+OPD_off;   /* delay in microns */
  delay_off +=  GaussNoise(rmsOPD);       /* OPD noise (independent) */
  delay_off = 1.0e-6 * delay_off / CLIGHT;/* convert to seconds */
  delay_off = delay_off + row->local_opd[0]; /* add modulation */

  /* add bandwidth decorrelation factor */
  arg = PI*freq_incr*delay_off;
  if (fabs(arg)>0.00001)
    bw_decorr = sin(arg)/arg;
  else
    bw_decorr = 1.0;
  corr_amp = amp *  bw_decorr;

  for (ireg=0; ireg<nreg; ireg++) {
    indx = 0; /* index into data array */
    for (j = 0; j<ny; j++) { /* loop over rows */
      for (i = 0; i<nx; i++) { /* loop over columns */
	bgcorr = background;
	/* determine frequency */
	freq = freq0 + (i-ref_pix)*freq_incr;
	/* measurement adding background, noise and source */
	phase = TWOPI * delay_off * freq + ireg * PI;
	value = bg[i]*background + GaussNoise(noise) + 
	  corr_amp * gain[i] * cos (phase);
	/* make sure the range isn't blown */
	testfact = fabs(value*DetGain)/32767.0;
	if (testfact>1.0) {
	  badrange = 1;
	  badfact = testfact;
	}
	row->data[ireg][indx++] = (DataType)(value*DetGain);
      } /* end loop over columns */
    } /* end loop over rows */
  } /* end loop over region */

  /* warning if called for - don't be excessive */
  if ((badrange) && (badfact>max_badfact)) {
    fprintf (stderr,"ERROR: Pixel vaues to large by factor %f\n",badfact);
    max_badfact = badfact;
  }
} /* end simulate_data */

float GaussNoise (float sigma)
/*----------------------------------------------------------------------- */
/*  Generate zero mean Gaussian noise (not very sophisticated)            */
/*  Must be initialized before first use                                  */
/*   Input:                                                               */
/*      sigma  Target sigma for the noise, if < 0.0 initialize only       */
/*   Returns:                                                             */
/*      a randum number                                                   */
/*----------------------------------------------------------------------- */
{
  float out;
  double sum;
  int i, n=12;
  time_t targ;

  /* initialize randum number generator if necessaey */
  if (sigma<0.0) {
    srand ((unsigned int)time(&targ));
    return sigma;
  }

  /* Blatant appeal to the central limit theorem */
  sum = 0.0;
  for (i=0; i<n; i++) sum += rand();
  sum = (sum  - (n/2 * (RAND_MAX+1.0))) / (RAND_MAX+1.0);
  out = (float)(sigma * sum);
  return out; /* done */
} /* end GaussNoise */

FILE* OPDfile_open (char *OPDfile, double tau0)
/*----------------------------------------------------------------------- */
/*  Open an OPD model file                                                */
/*  An OPD model file contains a series of OPD delays.                    */
/*  Comment lines start with "!".                                         */
/*  Each (noncomment) line contains a time in sec and a delay in microns  */
/*  as floats.   Sequence starts at 0.0 time                              */
/*     If the first non comment line starts with "TAU_0 =", then the      */
/*  following value is assumed to the the Tau_0 of the file.              */
/*  In this case the output OPD values will be scaled to the input        */
/*  value of tau0.                                                        */
/*   Input:                                                               */
/*      OPD file   Name of OPDfile                                        */
/*      tau0       Desired tau_0 of output, <=0 -> ignore                 */
/*   Returns:                                                             */
/*      File I/O pointer                                                  */
/*----------------------------------------------------------------------- */
{
  char szLine[121], *jerr;
  int iRead, i;
  float temp = 0.0, file_tau0;
  FILE *hFile = NULL;
  
  if (OPDfile==NULL) return hFile;  /* file name given */
  
  IamOK = 1; /* hope for the best */
  
  hFile = fopen (OPDfile, "rt");
  if (hFile==NULL) {
    fprintf(stderr,"ERROR opening OPD model file %s \n", OPDfile);
    fprintf(stderr,"CONTINUING WITH ZERO OPD \n");
    IamOK = 0;  /* not feeling so well */
    return hFile;
  }

  /* look for tau_0 in file */
  jerr = fgets (szLine, 120, hFile);
  while (szLine[0]=='!') jerr = fgets (szLine, 120, hFile);
  if (jerr==NULL) { /* error */
    fprintf(stderr,"ERROR reading OPD model file \n");
    fprintf(stderr,"CONTINUING WITH ZERO OPD \n");
    fclose(hFile);
    IamOK = 0;  /* not feeling so well */
    return NULL;
  } 
  if (!strncmp(szLine, "TAU_0 =", 7)) {
    iRead = sscanf (&szLine[7], "%f", &file_tau0);
    if (iRead!=1) {
      fprintf(stderr,"ERROR parsing tau0 from OPD model file line %s \n",
	      szLine);
      fprintf(stderr,"CONTINUING WITH ZERO OPD \n");
      fclose(hFile);
      IamOK = 0;  /* not feeling so well */
      return NULL;
    }
    /* if both valid, compute scaling */
    if ((file_tau0>0.0) && (tau0>0.0)) 
      /* have to scale by 5/6 power of ratio */
      OPDscale = pow (file_tau0/tau0, 0.83333);
    /* debug */
    /*fprintf(stderr,"OPD scaling %f \n",OPDscale);*/
  }
  
  /* read first three times, sequence taken to start at second */
  for (i=0; i<3; i++) {
    OPDfile_next (hFile,  &OPDtime[i], &OPDdelay[i]);
  }
  
  /* it all looks OK */
  return hFile;
} /* end OPFfile_open */
  
void OPDfile_next (FILE *hFile, double *_time, float *delay)
/*----------------------------------------------------------------------- */
/*  Read the next delay from the OPD sequence.                            */
/*  OPDfile_open must have been called first                              */
/*   Input:                                                               */
/*      hFile   File I/O pointer from OPDfile_open                        */
/*   Output:                                                              */
/*      time    time tag of next entry (sec)                              */
/*      delay   delay of next entry (micron)                              */
/*----------------------------------------------------------------------- */
{
  char szLine[121], *jerr;
  int iRead;
  float delay_out=0.0, time_out=0.0;

  *_time = 1.0e9;
  *delay = 0.0;

  if (hFile == NULL) return;  /* sanity check */
  if (!IamOK) return;         /* Is there some problem */

  jerr = fgets (szLine, 120, hFile);
  while (szLine[0]=='!') jerr = fgets (szLine, 120, hFile);
  if (jerr==NULL) { /* error */
    fprintf(stderr,"ERROR reading OPD model file \n");
    fprintf(stderr,"CONTINUING WITH ZERO OPD \n");
    fclose(hFile);
    IamOK = 0;  /* not feeling so well */
    return;
  } 
  iRead = sscanf (szLine, "%f %f", &time_out, &delay_out);
  if (iRead!=2) {
    fprintf(stderr,"ERROR parsing OPD model file line %s \n", szLine);
    fprintf(stderr,"CONTINUING WITH ZERO OPD \n");
    fclose(hFile);
    IamOK = 0;  /* not feeling so well */
    return;
  }

  /* set return values */
  *_time = time_out;
  *delay = delay_out * OPDscale;  /* scale for tau_0 */
  return; /* done */
}  /* OPDfile_next */

void OPDfile_find (FILE *hFile, double _time)
/*----------------------------------------------------------------------- */
/*  Position the global array to so the the requested time is within the  */
/*  range covered in the global arrays                                    */
/*   Input:                                                               */
/*      hFile   File I/O pointer from OPDfile_open                        */
/*      time    The target time in sec                                    */
/*----------------------------------------------------------------------- */
{

  if (hFile == NULL) return;  /* sanity check */
  if (!IamOK) return;         /* Is there some problem */

  /* can't go backwards so assume it's OK if time is less than the last */
  if (_time<OPDtime[2]) return;

  /* shuffle till happy */
  while ((_time>=OPDtime[2]) && (IamOK)) {
    /* shuffle */
    OPDdelay[0] = OPDdelay[1];
    OPDtime[0]  = OPDtime[1];
    OPDdelay[1] = OPDdelay[2];
    OPDtime[1]  = OPDtime[2];
    /* read next */
    OPDfile_next (hFile, &OPDtime[2], &OPDdelay[2]);
  }
}  /* OPDfile_find */

float OPDfile_read (FILE *hFile, double _time)
/*----------------------------------------------------------------------- */
/*  Read/interpolate the next delay from the sequence.                    */
/*  OPDfile_open must have been called first                              */
/*   Input:                                                               */
/*      hFile   File I/O pointer from OPDfile_open                        */
/*      time    Sequence time in seconds for the OPD value                */
/*   Returns:                                                             */
/*      The next OPD in the sequence (microns), 0.0 if something is wrong */
/*----------------------------------------------------------------------- */
{
  float sum, sumwt, wt, val, x, incr, out=0.0;
  int i, n=3, ihalf=1;

  if (hFile == NULL) return out;  /* sanity check */
  if (!IamOK) return out;         /* Is there some problem */

  OPDfile_find (hFile, _time); /* align arrays */
  if (!IamOK) return out;     /* Was there some problem */

  /* get interpolation kernal */
  incr = (OPDtime[1] - OPDtime[0]);
  if (incr<0.0) incr = 1.0;
  x = (_time - OPDtime[0]) / incr; /* location in pixels */
  iminfn (x, n, ihalf, &OPDintr[0], &OPDcen);

  /* interpolate */ 
  sum = sumwt = 0.0;
  for (i=0; i<3; i++) {
    wt = OPDintr[i+1];
    val = OPDdelay[OPDcen+i];
    sum += wt*val;
    sumwt += wt;
  }

  if (sumwt>0.0) out = sum / sumwt; /* get value */

  return out; /* done */
} /* end OPDfile_read  */

void  OPDfile_close(FILE* hFile)
/*----------------------------------------------------------------------- */
/*  Close an OPDfile Opened with OPDfile_open                             */
/*   Input:                                                               */
/*      hFile   File I/O pointer from OPDfile_open                        */
/*----------------------------------------------------------------------- */
{
  if (hFile == NULL) return;  /* sanity check */
  fclose (hFile);

  IamOK = 0;  /* I've gone offline */

} /* end  OPDfile_close */
/* */
